@@ -230,7 +230,7 @@ namespace snmalloc
       auto ss = address_cast(p) + SUPERSLAB_SIZE;
       for (size_t i = 0; i < size_bits - SUPERSLAB_BITS; i++)
       {
-        size_t run = 1ULL << i;
+        size_t run = bits::one_at_bit(i);
         PagemapProvider::pagemap().set_range(
           ss, static_cast<uint8_t>(CMLargeRangeMin + i), run);
         ss = ss + SUPERSLAB_SIZE * run;
