@@ -57,7 +57,7 @@ namespace snmalloc
   // global state of the allocator.  This is currently stored in the memory
   // provider, so we add this in.
   template<SNMALLOC_CONCEPT(ConceptPAL) PAL>
-  class MemoryProviderStateMixin : public PalNotificationObject
+  class MemoryProviderStateMixin
   {
     /**
      * Simple flag for checking if another instance of lazy-decommit is
@@ -136,36 +136,27 @@ namespace snmalloc
     static MemoryProviderStateMixin<PAL>* make() noexcept
     {
       // Temporary stack-based storage to start the allocator in.
-      MemoryProviderStateMixin<PAL> local{};
+      AddressSpaceManager<PAL> local{};
 
       // Allocate permanent storage for the allocator usung temporary allocator
       MemoryProviderStateMixin<PAL>* allocated =
-        local.alloc_chunk<MemoryProviderStateMixin<PAL>, 1>();
+        reinterpret_cast<MemoryProviderStateMixin<PAL>*>(
+          local.template reserve_with_left_over<true>(
+            sizeof(MemoryProviderStateMixin<PAL>)));
 
       if (allocated == nullptr)
         error("Failed to initialise system!");
 
-#ifdef GCC_VERSION_EIGHT_PLUS
-#  pragma GCC diagnostic push
-#  pragma GCC diagnostic ignored "-Wclass-memaccess"
-#endif
-      // Put temporary allocator we have used, into the permanent storage.
-      // memcpy is safe as this is entirely single threaded: the move
-      // constructors were removed as unsafe to move std::atomic in a
-      // concurrent setting.
-      ::memcpy(
-        &(allocated->address_space),
-        &(local.address_space),
-        sizeof(AddressSpaceManager<PAL>));
-#ifdef GCC_VERSION_EIGHT_PLUS
-#  pragma GCC diagnostic pop
-#endif
+      // Move address range inside itself
+      allocated->address_space = std::move(local);
 
       // Register this allocator for low-memory call-backs
       if constexpr (pal_supports<LowMemoryNotification, PAL>)
       {
-        allocated->PalNotificationObject::pal_notify = &(allocated->process);
-        PAL::register_for_low_memory_callback(allocated);
+        auto callback =
+          allocated->template alloc_chunk<LowMemoryNotificationObject, 1>(
+            allocated);
+        PAL::register_for_low_memory_callback(callback);
       }
 
       return allocated;
@@ -218,15 +209,26 @@ namespace snmalloc
       lazy_decommit_guard.clear();
     }
 
-    /***
-     * Method for callback object to perform lazy decommit.
-     */
-    static void process(PalNotificationObject* p)
+    class LowMemoryNotificationObject : public PalNotificationObject
     {
-      // Unsafe downcast here. Don't want vtable and RTTI.
-      auto self = reinterpret_cast<MemoryProviderStateMixin<PAL>*>(p);
-      self->lazy_decommit();
-    }
+      MemoryProviderStateMixin<PAL>* memory_provider;
+
+      /***
+       * Method for callback object to perform lazy decommit.
+       */
+      static void process(PalNotificationObject* p)
+      {
+        // Unsafe downcast here. Don't want vtable and RTTI.
+        auto self = reinterpret_cast<LowMemoryNotificationObject*>(p);
+        self->memory_provider->lazy_decommit();
+      }
+
+    public:
+      LowMemoryNotificationObject(
+        MemoryProviderStateMixin<PAL>* memory_provider)
+      : PalNotificationObject(&process), memory_provider(memory_provider)
+      {}
+    };
 
   public:
     /**
@@ -238,8 +240,8 @@ namespace snmalloc
     {
       // Cache line align
       size_t size = bits::align_up(sizeof(T), 64);
-      size = bits::next_pow2(bits::max(size, alignment));
-      void* p = address_space.template reserve<true>(size);
+      size = bits::max(size, alignment);
+      void* p = address_space.template reserve_with_left_over<true>(size);
       if (p == nullptr)
         return nullptr;
 
