@@ -30,7 +30,7 @@ int main()
   // It is also large enough for the example to run in.
   // For 1MiB superslabs, SUPERSLAB_BITS + 4 is not big enough for the example.
   size_t large_class = 28 - SUPERSLAB_BITS;
-  size_t size = 1ULL << (SUPERSLAB_BITS + large_class);
+  size_t size = bits::one_at_bit(SUPERSLAB_BITS + large_class);
   void* oe_base = mp.reserve<true>(large_class);
   void* oe_end = (uint8_t*)oe_base + size;
   PALOpenEnclave::setup_initial_range(oe_base, oe_end);
